# -*- coding: utf-8 -*-
#
# Register Windows Systems at OPSI
#
# Copyright 2015 DIGITEC GmbH based on work on Univention Copyright 2012 Univention GmbH
#
# http://www.digitec.de/
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

name = 'opsilistener'
description = 'listener to add, remove and update windows clients in OPSI'
filter = '(objectClass=univentionWindows)'
attributes = ['macAddress', 'aRecord', 'cn', 'univentionInventoryNumber', 'description']
modrdn = '1'

RENAME = False

__package__=''  # workaround for PEP 366

import simplejson as json  # the "normal" json module doesn't work in optimize mode, no idea why
from subprocess import Popen, PIPE
import traceback
import listener
from univention.debug import debug, LISTENER, INFO, ERROR, WARN
import univention.uldap

filter = listener.baseConfig.get('digitec/opsi/listener/host/filter', filter)


ldap_info = {}
def opsi_admin(*args, **kwargs):
	""" Run opsi-admin -d method with any number of arguments.

	:args: All arguments are passed to opsi-admin -d method
	:returns: Return opsi-admin process object
	"""
	assert args

	cmd = ['/usr/bin/opsi-admin', '-d', 'method'] + list(args)
	listener.setuid(0)
	# HOME has to be defined to work around a bug in opsi-admin
	p = Popen(cmd, env={'HOME': '/root', \
		'LANG': listener.baseConfig.get('locale', 'C').split(':')[0]}, \
			**kwargs)
	data = p.communicate()[0]

	listener.unsetuid()
	# rc = listener.run('/usr/bin/opsi-admin', cmd, uid=0)
	if p.returncode != 0:
		debug(LISTENER, ERROR, '%s: opsi-admin command terminated with error code %d: %s' %
				(name, p.returncode, ' '.join(cmd)))
		return None
	return data


def get_hostobject(hostid, ignore_errors=False):
	"""Retrieve JSON data representing hostid

	:hostid: FQDN of the requested host
	:returns: None or JSON data representing hostid
	"""

	try:
        # query opsi for existing information about client
		data = opsi_admin('host_getObjects', '[]', '{"id": "%s"}' % hostid, \
			stdout=PIPE)
		return json.loads(data)[0]
	except ValueError:
		if not ignore_errors:
			debug(LISTENER, ERROR, '%s: opsi-admin data: %s' % (name, data))
			debug(LISTENER, ERROR, traceback.format_exc())

def get_all_hostobjects():
	"""Retrieve  JSON data representing all hosts

	:returns: None or JSON data representing hosts
	"""

	# query opsi for existing information about client
	try:
		data = opsi_admin('host_getObjects', '[]', '{"type":"OpsiClient"}',  \
			stdout=PIPE)
		return json.loads(data)
	except ValueError:
		debug(LISTENER, ERROR, '%s: opsi-admin data: %s' % (name, data))
		debug(LISTENER, ERROR, traceback.format_exc())

def update_host(hostid, new, old, host_obj=None):
	"""Update an OPSI host object

	:hostid: FQDN of the requested OPSI host
	:new: dictionary as it is returned by LDAP search
	:old: dictionary as it is returned by LDAP search
	:host_obj: optional host object as it is returned by get_hostobject
	:returns: returns result of opsi_admin call
	"""

	debug(LISTENER, INFO, '%s: updating OPSI client: %s' % (name, hostid))

	mapping_ldap_opsi = { \
			'aRecord':                    'ipAddress', \
			'macAddress':                 'hardwareAddress', \
			'description':                'description', \
			'univentionInventoryNumber':  'inventoryNumber'}

	if not host_obj:
		host_obj = get_hostobject(hostid.lower())

	if not host_obj:
		debug(LISTENER, ERROR, '%s: stopping, unable to find OPSI client: %s' % \
				(name, hostid))
		return

	changed = 0
	for i in ('aRecord', 'description', \
			'univentionInventoryNumber'):
		attr_new = sorted(new.get(i, [None]))[0]
		attr_old = sorted(old.get(i, [None]))[0]
		if attr_new != attr_old:
			debug(LISTENER, INFO, '%s: updating OPSI attribute %s for host %s to %s' % \
					(name, mapping_ldap_opsi[i], hostid, attr_new))
			if attr_new:
				# encode string to allow OPSI to decode it properly
				attr_new = attr_new.encode()
			host_obj[mapping_ldap_opsi[i]] = attr_new
			changed = 1
		else:
			# update all attributes from LDAP to create a sync state
			if attr_new:
				# encode string to allow OPSI to decode it properly
				attr_new = attr_new.encode()
			host_obj[mapping_ldap_opsi[i]] = attr_new

	# sync mac: test if the mac in opsi is in the ucs ldap.
	changed_mac=1
	for ucs_mac in new.get('macAddress', [None]):
	        if ucs_mac.lower() == host_obj['hardwareAddress'].lower():
			debug(LISTENER, INFO, '%s: found opsi mac %s in ucs no update for host %s.' % \
				      (name, host_obj['hardwareAddress'], hostid ))
			changed_mac=0

	if changed_mac:
		new_mac = sorted(new.get('macAddress', [None]))[0]
		debug(LISTENER, INFO, '%s: updating OPSI hardwareAddresse from %s to %s for host %s.' % \
					(name, host_obj['hardwareAddress'], new_mac, hostid))
		changed=1		
		host_obj['hardwareAddress'] = new_mac
		
	if changed:
		# WORKAROUND: host_updateObjects doesn't work when setting
		# attributes like ipAddress and hardwareAddress to null, therefore
		# use host_createObjects. By pulling the object first via
		# opsi-admin and updating it on the way, last seen and created
		# timestamps are kept
		if listener.baseConfig.is_true('digitec/opsi/listener/host/modify',True):
		   opsi_admin("host_createObjects", json.dumps(host_obj))
		else:
                   debug(LISTENER, INFO, '%s: modify OPSI client: %s skipped' % (name, hostid))
	else:
		debug(LISTENER, INFO, '%s: no changes found for OPSI client: %s' % (name, hostid))

def delete_host(hostid):
	if listener.baseConfig.is_true('digitec/opsi/listener/host/delete',True):
	    debug(LISTENER, INFO, '%s: deleting OPSI client: %s' % (name, hostid))
	    opsi_admin("host_delete", hostid)
        else:
	    debug(LISTENER, INFO, '%s: deleting of OPSI client skipped: %s' % (name, hostid))
   

def create_host(hostid,desc,mac,ip,inventory_nr):
	if listener.baseConfig.is_true('digitec/opsi/listener/host/create',True):
            debug(LISTENER, INFO, '%s: creating OPSI client: %s' % (name, hostid))
	    opsi_admin("host_createOpsiClient", hostid,
		   "null", desc, "null", mac, ip, inventory_nr)
	else:
	    debug(LISTENER, INFO, '%s: creating of OPSI client skipped: %s' % (name, hostid))

def ldap_reconnect():
        debug(LISTENER, INFO, '%s: ldap reconnect triggered' % name)
        if 'ldapserver' in ldap_info and 'basedn' in ldap_info and 'binddn' in ldap_info and 'bindpw' in ldap_info:
                try:
                        ldap_info['lo'] = univention.uldap.access(host = ldap_info['ldapserver'],
                                                                  base = ldap_info['basedn'],
                                                                  binddn = ldap_info['binddn'],
                                                                  bindpw = ldap_info['bindpw'],
                                                                  start_tls = 2)
                except ValueError, ex:
                        debug(LISTENER, ERROR, '%s: ldap reconnect failed: %s' % (name,str(ex)))
                        ldap_info['lo'] = None
                else:
                        if ldap_info['lo'] == None:
                                debug(LISTENER, ERROR, '%s: ldap reconnect failed' % name)

def setdata(key, value):
        if key == 'bindpw':
                debug(LISTENER, INFO, '%s: listener passed key="%s" value="<HIDDEN>"' % (name,key))
        else:
                debug(LISTENER, INFO, '%s: listener passed key="%s" value="%s"' % (name,key, value))

        if key in [ 'ldapserver', 'basedn', 'binddn', 'bindpw' ]:
                ldap_info[ key ] = value
        else:
                debug(LISTENER, INFO, '%s: listener passed unknown data (key="%s" value="%s")' % (name,key, value))

        if key == 'ldapserver':
                debug(LISTENER, INFO, '%s: ldap server changed to %s' % (name,value))
                ldap_reconnect()


def initialize():
	debug(LISTENER, INFO, '%s: Initialize' % name)
	if not ldap_info['lo']:
		ldap_reconnect()
	if ldap_info['lo']:		
 	    hosts = get_all_hostobjects()
	    for host in hosts:
                hostid=host['id'].lower()
		debug(LISTENER,INFO,'%s: Found OPSI-Host: %s' % (name,hostid))
		domainname = listener.baseConfig.get('domainname', '')
		cn=hostid.replace(".%s"%domainname,"")
		search='(&%s(cn=%s))' % (filter, cn)
		debug(LISTENER,INFO,'%s: Search OPSI-Host: %s' % (name,search))
                if not ldap_info['lo'].search(filter=search):
		    delete_host(hostid)

def handler(dn, new, old, command):
	global RENAME
	domainname = listener.baseConfig.get('domainname', '')
	# object modified
	if command == 'm':
		if new and old:
			cn = new['cn'][0]
			update_host('.'.join((cn, domainname)), new, old)
		else:
			debug(LISTENER, ERROR, '%s: Modify not possible, (old and new == False). new: %r, old: %r' % (name, new, old))

	# object deleted
	elif command == 'd':
		if old:
			cn = old['cn'][0]
			hostid = '.'.join((cn, domainname))

			delete_host(hostid)
		else:
			debug(LISTENER, ERROR, '%s: Delete not possible, (old == False). old: %r' % (name, old))

	# object created
	elif command == 'a':
		if RENAME:
			debug(LISTENER, INFO, '%s: Detected rename of client, destination DN: %s' % (name, dn))
			RENAME = False
		elif new:
			cn = new['cn'][0]
			hostid = '.'.join((cn, domainname))
			hostid = hostid.lower()
			host_obj = get_hostobject(hostid, ignore_errors=True)

			if host_obj:
				# update client instead of creating a new one, this will prevent
				# the creation of a new OPSI host keyelif new:
				update_host(hostid, new, {}, host_obj=host_obj)

			else:
				# create a new client, this will also the creation of a new OPSI
				# host key, even if the object already exists
				ip = sorted(new.get('aRecord', ['null']))[0]
				mac = sorted(new.get('macAddress', ['null']))[0]
				desc = sorted(new.get('description', ['null']))[0]
				inventory_nr = sorted(new.get('univentionInventoryNumber', ['null']))[0]
				create_host(hostid,desc,mac,ip,inventory_nr)
		else:
			debug(LISTENER, ERROR, '%s: Create not possible, (new == False). new: %r' % (name, new))
	# Object Move
	elif command == 'r':
		debug(LISTENER, INFO, '%s: Detected rename of client, source DN: %s' % (name, dn))
		RENAME = True
	
